/** mergepipe.c   simple pipelined merge sort     C. Kessler 09/99
 *                binary tree of merges for floatingpoint streams
 *                as a demo for asynchronous message passing
 */
#include <mpi.h>

#define StreamBufSize 4
#define EndOfStream 99999.0   // some large value as end marker
#define N 80
int P;
sh float usrt[N];
sh float *psrt[4096];  // one slice for each processor
    
int fcmp( void *a, void *b )  // for floats
{
 if      (*((float *)a) < *((float *)b)) return -1;
 else if (*((float *)a) > *((float *)b)) return 1;
 else                                    return 0;
}

void flushStreamBuf( float *stream, int dest )
{
 pprintf("flushStream to %d: %d %d %d %d\n", dest,
          (int)stream[0], (int)stream[1], (int)stream[2], (int)stream[3] );
 if (dest >= 1) 
    MPI_Send( stream, StreamBufSize, MPI_FLOAT, dest, 0, MPI_COMM_WORLD );
 else { // root node:
    int i;
    for (i=0; i<StreamBufSize && stream[i]!=EndOfStream; i++)
       printf("[%2.0f]", stream[i] );
    printf("\n");
 }
}

void readStreamBuf( float *stream, int source )
{
 MPI_Status status;
 pprintf("readStream from %d...\n", source );
 if (source < P )
    MPI_Recv( stream, StreamBufSize, MPI_FLOAT, source, 
              MPI_ANY_TAG, MPI_COMM_WORLD, &status );
 else { // leaf node:
    int k = source - P;
    memcpy( stream, psrt[k], StreamBufSize*sizeof(float) );
    psrt[k] += StreamBufSize;
 }
 pprintf("readStream from %d: got %d %d %d %d\n", source,
          (int)stream[0], (int)stream[1], (int)stream[2], (int)stream[3] );
}

void mergeStreams( int myLeftChild, int myRightChild, int myParent )
{
 float left[StreamBufSize], right[StreamBufSize],
       out[StreamBufSize]; 
 int leftpos = 0, rightpos = 0, outpos = 0;
 readStreamBuf( left, myLeftChild );
 readStreamBuf( right, myRightChild );
 while (!(  left[leftpos]==EndOfStream
         && right[rightpos]==EndOfStream )) 
 {  // at entry: at least one item in left and right
    if (left[leftpos]==EndOfStream)
       out[outpos++] = right[rightpos++];
    else
    if (right[rightpos]==EndOfStream)
       out[outpos++] = left[leftpos++];
    else
    if (fcmp(left+leftpos, right+rightpos) <= 0)
       out[outpos++] = left[leftpos++];
    else
       out[outpos++] = right[rightpos++];
    if (outpos == StreamBufSize) {
       flushStreamBuf( out, myParent );
       outpos = 0;
    }
    if (leftpos == StreamBufSize) {
       readStreamBuf( left, myLeftChild );
       leftpos = 0;
    }
    if (rightpos == StreamBufSize) {
       readStreamBuf( right, myRightChild );
       rightpos = 0;
    }
 } // while
 out[outpos] = EndOfStream;
 flushStreamBuf( out, myParent ); 
}

void main( void )
{
  MPI_Comm comm;
  int myid, i, slice, mysize;
  
  MPI_Init(NULL, NULL);

  MPI_Comm_rank( MPI_COMM_WORLD, &myid );
  MPI_Comm_size( MPI_COMM_WORLD, &P );
  
  // initialize unsorted array:
  forall (i,0,N,P)
     usrt[i] = (float)((N-i)*i*(i+1) % 1024) + 1.0;
  
  // sort into P buckets psrt[0]...psrt[p-1]:
  slice = (N+P-1) / P;   // == ceil(N/P)
  if (myid*slice >= N)  mysize = 0;
  else if (myid*slice >= N-slice) mysize = N - (myid*slice);
  else mysize = slice;
  qsort( usrt+myid*slice, mysize, sizeof(float), fcmp );
  psrt[myid] = (float *)shmalloc((mysize+StreamBufSize+1)*sizeof(float)); 
  memcpy( psrt[myid], usrt+myid*slice, mysize*sizeof(float) );
  psrt[myid][mysize] = EndOfStream;
 
  MPI_Barrier( MPI_COMM_WORLD );
  if (myid==0) {
     for (i=0; i<N; i++)
     printf("%2.2f ", usrt[i] );
     printf("\n");
  }
  MPI_Barrier( MPI_COMM_WORLD );

  pprintf("merging\n");
  if (myid > 0)
     mergeStreams( 2*myid, 2*myid+1, myid/2 );
  
  MPI_Finalize();
  barrier;
}
