/*  pipe.c   by C.W.Kessler 05/95 updated 07/99
 *  
 *  pipelined matrix-vector product implementation in Fork95
 *
 *  n must be a power of 2 and at least be p/2.
 *  p must be a power of 2 minus 1.
 *
 *  Condition: group-relative Processor ID's '$' must be consecutively
 *             numbered from 0 to groupsize - 1.
 */

#include <fork.h>
#include <syscall.h>
#include <stdlib.h>
#include <io.h>

// #define max(a,b) ((a)>=(b)?(a):(b))   // defined in math.h


/* Pipeline graph consisting of p nodes: */

typedef struct pipenode {
  int nodeid;
  int npreds;
  struct pipenode **pred;
  int stage;
  void *data;
} *PipeNode;

typedef struct graph {
  int n;              // number of nodes
  PipeNode *nodes;    // array of pipeline nodes
  int depth;
} *Graph;


typedef struct channel {
  float accsum;    // accumulated sum
  float xdown;     // location to pass x-values down
  int   xread;     // flag indicating x-value has been read
} *Channel;

sync Graph linear_pipeline( int n ) 
{
  sh Channel *ch;
  sh int p = 0;
  sh Graph pipeline;
  sh PipeNode *pipenodes;
  int i;
  $ = mpadd( &p, 1 );
  pipeline = (Graph) shalloc( sizeof(struct graph));
  pipenodes = (PipeNode *) shalloc( n * sizeof( PipeNode ) );
  pipeline->n = n;
  pipeline->nodes = pipenodes;
  pipeline->depth = n;
  ch = (Channel *)shalloc( n*sizeof(Channel));
  farm
   forall(i, 0, n, p ) {
     pipenodes[i] = shmalloc( sizeof(struct pipenode));
     pipenodes[i]->pred = (PipeNode *)shmalloc( 1 * sizeof(PipeNode) );
     pipenodes[i]->nodeid = i;
     if (i==0) {
       pipenodes[i]->npreds = 0;
       pipenodes[i]->pred[0] = NULL;  // no child
     }
     else {
       pipenodes[i]->npreds = 1;
       pipenodes[i]->pred[0] = pipenodes[i-1];
     }
     pipenodes[i]->stage = i;
     ch[i] = (Channel)shmalloc(sizeof(struct channel));
     ch[i]->accsum = 0.0;
     ch[i]->xdown = 0.0;
     ch[i]->xread = 0;
     pipeline->nodes[i]->data = ch[i];
   }
  return pipeline;
}


typedef struct inputset {
  int n, m;
  float **skewedmatrix;
  float *paddedvector;
} *InputSet;


/** executed simultaneously by independent groups, 
 *  one group for each node of G.
 */
sync void linpstep( sh Graph G, sh PipeNode mynode, sh int t,
                    sh void *A, sh void *R )
{
 sh int mynodeid = mynode->nodeid;
 sh float **a = ((InputSet)A)->skewedmatrix;
 sh float *x = ((InputSet)A)->paddedvector;
 sh float m = ((InputSet)A)->m;
 sh float xval, aval, yval, oldyval;

 seq pprintf("linpstep %d %p %p n=%d m=%d\n", t, A, R, ((InputSet)A)->n,
             ((InputSet)A)->m );
 // read input vector element:
 if (mynode->stage == 0) {
    xval = x[t];
 }
 else {
    xval = ((Channel)(mynode->pred[0]->data))->xdown;
    ((Channel)(mynode->pred[0]->data))->xread = 1;
 }

 yval = ((Channel)(mynode->data))->accsum;
 aval = a[mynodeid][t];
 yval += aval * xval;
 ((Channel)(mynode->data))->accsum = yval;
 seq pprintf("update %d: %d=%d*%d+y\n", mynode->nodeid, (int) yval,(int)aval,(int)xval );

 // forward xval to bottom processor:
 if (t > mynodeid)   // next level of processors is active:
   if (mynode->stage < G->depth-1)
     while (! ((Channel)(mynode->data))->xread)  ;
     // wait until P(mynodeid+1) has read data
 ((Channel)(mynode->data))->xdown = xval;

 // write result when valid:
 if (t==m+mynodeid-1)
    ((float *)R)[mynodeid] = yval; 
}


sync void Pipe( sh Graph pipeline, sh int n,
                sh void *input, sh int niter, sh int elsize,
                sh void *result,
                sh sync void (*step)( sh Graph, sh PipeNode, sh int,
                                      sh void *, sh void *),
                sh sync void (*seqalg)( sh int, sh void *, sh void *))
{
 sh int t;
 if (groupsize() < n) {
    seqalg( n, input, result );
    return;
 }
 for (t=0; t<niter; t++) 
   fork (n; @=$%n; $=$/n )
     if (t >= pipeline->nodes[@]->stage) {
       seq pprintf("calling pipestep(%d)\n", t);
       step( pipeline, pipeline->nodes[@], t,
             (void **)input, (void **)result );
     }
}

sync void other_mv( sh int n, sh void *input, sh void *output )
{
  farm pprintf("Horst\n");
}

void main( void )
{
 start {
   sh int n = 4;
   sh int m = 5;
   sh int p = __STARTED_PROCS__;
   sh float **a, *b, *y;
   sh InputSet in;
   int i, j;

   // preset the input array:
   a = (float **) shalloc(n*sizeof(float *));
   farm forall(i,0,n,p)
           a[i] = (float *)shmalloc((m+i)*sizeof(float));
   farm
     forall(i,0,n,p) {
        for(j=0; j<i; j++) a[i][j] = 0.0;
        for(j=i; j<m+i; j++)  a[i][j] = (float)(i*10+(j-i));
     }
   b = (float *) shalloc((m+n-1)*sizeof(float));
   farm {
     forall(j,0,m,p)  b[j] = (float)j;
     forall(j,m,m+n-1,p)  b[j] = 0.0;
   }
   in = (InputSet)shalloc(sizeof(struct inputset));
   in->n = n;
   in->m = m;
   in->skewedmatrix = a;
   in->paddedvector = b;
   seq {
     prS("\nSource Array b:\n");
     for (i=0; i<m; i++) printf(" %f", b[i] );
     printf("\n");
   }
   y = (float *) shalloc(n*sizeof(float));

   Pipe( linear_pipeline( n ), n, in, n+m-1,
         sizeof(float), (void *)y, linpstep, other_mv );

   seq printf("\nResult: %f %f %f %f\n", y[0],y[1],y[2],y[3] );
  }
}
