/*************************************************/
/*                                               */
/*      mpi.h                                    */
/*      message passing interface in fork95      */
/*      v0.15                                    */
/*      13.09.1999 Christoph W. Kessler,         */
/*      Oliver Fritzen                           */
/*************************************************/

/* Results of the compare operations */
/* These should stay ordered */

#ifndef mpi_header 

#ifndef __FORK_HEADER__
#include <fork.h>
#endif

#define MPI_IDENT     0
#define MPI_CONGRUENT 1 
#define MPI_SIMILAR   2
#define MPI_UNEQUAL   3

/* Data types */

typedef int MPI_Datatype;

#define MPI_CHAR           ((MPI_Datatype)1)
#define MPI_UNSIGNED_CHAR  ((MPI_Datatype)2)
#define MPI_BYTE           ((MPI_Datatype)3)
#define MPI_SHORT          ((MPI_Datatype)4)
#define MPI_UNSIGNED_SHORT ((MPI_Datatype)5)
#define MPI_INT            ((MPI_Datatype)6)
#define MPI_UNSIGNED       ((MPI_Datatype)7)
#define MPI_LONG           ((MPI_Datatype)8)
#define MPI_UNSIGNED_LONG  ((MPI_Datatype)9)
#define MPI_FLOAT          ((MPI_Datatype)10)
#define MPI_DOUBLE         ((MPI_Datatype)11)
#define MPI_LONG_DOUBLE    ((MPI_Datatype)12)
#define MPI_LONG_LONG_INT  ((MPI_Datatype)13)

#define MPI_PACKED         ((MPI_Datatype)14)
#define MPI_LB             ((MPI_Datatype)15)
#define MPI_UB             ((MPI_Datatype)16)

#define MPI_FLOAT_INT      ((MPI_Datatype)17)
#define MPI_DOUBLE_INT     ((MPI_Datatype)18)
#define MPI_LONG_INT       ((MPI_Datatype)19)
#define MPI_SHORT_INT      ((MPI_Datatype)20)
#define MPI_2INT           ((MPI_Datatype)21)
#define MPI_LONG_DOUBLE_INT ((MPI_Datatype)22)

/*
 * error classes
 */
#define MPI_ERR_BUFFER          1       /* invalid buffer pointer */
#define MPI_ERR_COUNT           2       /* invalid count argument */
#define MPI_ERR_TYPE            3       /* invalid datatype argument */
#define MPI_ERR_TAG             4       /* invalid tag argument */
#define MPI_ERR_COMM            5       /* invalid communicator */
#define MPI_ERR_RANK            6       /* invalid rank */
#define MPI_ERR_REQUEST         7       /* invalid request handle */
#define MPI_ERR_ROOT            8       /* invalid root */
#define MPI_ERR_GROUP           9       /* invalid group */
#define MPI_ERR_OP              10      /* invalid operation */
#define MPI_ERR_TOPOLOGY        11      /* invalid topology */
#define MPI_ERR_DIMS            12      /* invalid dimension argument */
#define MPI_ERR_ARG             13      /* invalid argument */
#define MPI_ERR_UNKNOWN         14      /* unknown error */
#define MPI_ERR_TRUNCATE        15      /* message truncated on receive */
#define MPI_ERR_OTHER           16      /* FORK_MPI error */
#define MPI_ERR_INTERN          17      /* internal MPI error */
#define MPI_ERR_IN_STATUS       18      /* error code is in status */
#define MPI_ERR_PENDING         19      /* pending request */
#define MPI_ERR_SYSRESOURCE     20      /* out of system resources */
#define MPI_ERR_LOCALDEAD       21      /* process in local group is dead */
#define MPI_ERR_REMOTEDEAD      22      /* process in remote group is dead */
#define MPI_ERR_VALUE           23      /* truncated info value */
#define MPI_ERR_FLAGS           24      /* mismatched run-time flags */
#define MPI_ERR_SERVICE         25      /* name publishing service error */
#define MPI_ERR_NAME            26      /* name not published */
#define MPI_ERR_SPAWN           27      /* error while spawning processes */
#define MPI_ERR_KEYVAL          28      /* invalid key value */
#define MPI_ERR_INFO_NOKEY      29      /* no such info key */
#define MPI_ERR_WIN             30      /* invalid window */
#define MPI_ERR_EPOCH           31      /* mismatched one-sided synch. */
#define MPI_ERR_TYPENOTSUP      32      /* operation not supported on type */
#define MPI_ERR_INFO_KEY        33      /* invalid info key */
#define MPI_ERR_INFO_VALUE      34      /* invalid info value */
#define MPI_ERR_LASTCODE        35      /* last error code */




/* 
   The layouts for the types MPI_DOUBLE_INT etc are simply
   struct { 
   double var;
   int    loc;
   }
   */

typedef struct { 
  float var;
  int   loc;
} mpi_float_int;

typedef mpi_float_int mpi_double_int;
typedef mpi_float_int mpi_long_double_int;

typedef struct { 
  long var;
  int  loc;
} mpi_2int;

typedef mpi_2int mpi_short_int;
typedef mpi_2int mpi_long_int;
   


#define SYSTEM_WORD_LENGTH 32
/* SYSTEM_WORD_LENGTH is the length of a word in bits */
/* (and, since SB-PRAM-words are 4 Byte SWL is 32 */

#define MPI_NR_OF_RED_FUNCS 32
/* Max. number of binary functions used by MPI_Reduce */
/* 12 are predefined, up to 20 can be user-defined */
/* (if you want more, change the above constant) */
	  

/* Groups */

typedef struct {
  int size;
  int *members;
  int *ranks;
} MPI_Group;

/* Communicators */

typedef struct {
  MPI_Group group;
  sh int *comm_barrier;
} MPI_Comm;

/* parallel operations */

typedef int MPI_Op; 

#define MPI_MAX     0
#define MPI_MIN     1
#define MPI_SUM     2
#define MPI_PROD    3
#define MPI_LAND    4
#define MPI_BAND    5
#define MPI_BOR     6
#define MPI_BXOR    7
#define MPI_LOR     8
#define MPI_LXOR    9
#define MPI_MAXLOC 10
#define MPI_MINLOC 11

/* MAX values */
#define MPI_MAX_PROCESSOR_NAME 256
#define MPI_MAX_ERROR_STRING   512
#define MPI_MAX_NAME_STRING     63      /* How long a name do you need ? */

/* Pre-defined constants */

#define MAILBOX_SIZE 100
#define MPI_UNDEFINED      (-32766)
#define MPI_UNDEFINED_RANK MPI_UNDEFINED
#define MPI_KEYVAL_INVALID 0

/* source wildcard alias */

#define MPI_ANY_SOURCE -1


/* tag wildcard alias */
#define MPI_ANY_TAG -1

/* 
 * Status object.  It is the only user-visible MPI data-structure 
 */

typedef struct { 
  int count;
  int MPI_SOURCE;
  int MPI_TAG;
  int MPI_ERROR;
  int private_count;
} MPI_Status;

extern pr MPI_Comm MPI_COMM_WORLD;
extern pr MPI_Group MPI_GROUP_EMPTY;

/* MPI functions (at least some of them) */

/* init and finalize */
extern int MPI_Init(int *, char ***);
extern int MPI_Finalize(void);

/* (some) communication functions */
extern int MPI_Send(const void *, int, MPI_Datatype, int, int, MPI_Comm);
extern int MPI_Bcast(const void *, int, MPI_Datatype, int, MPI_Comm);
extern int MPI_Recv(const void *, int, MPI_Datatype, int, int, MPI_Comm, MPI_Status *);
extern int MPI_Gather(void *sendbuf, int sendcount, MPI_Datatype sendtype,
		      void *recvbuf, int recvcount, MPI_Datatype recvtype, 
		      int root, MPI_Comm comm);
extern int MPI_Scatter(void *sendbuf, int sendcount, MPI_Datatype sendtype,
                      void *recvbuf, int recvcount, MPI_Datatype recvtype, 
                      int root, MPI_Comm comm);
extern int MPI_Reduce(void *sendbuf, void *recvbuf, int count,
		      MPI_Datatype datatype, MPI_Op op, int root,
		      MPI_Comm comm);
extern int MPI_Barrier(MPI_Comm comm);

/* Comm functions */
extern int MPI_Comm_group(MPI_Comm, MPI_Group *);
extern int MPI_Comm_create(MPI_Comm, MPI_Group, MPI_Comm *);
extern int MPI_Comm_free(MPI_Comm *);
extern int MPI_Comm_size(MPI_Comm, int *);
extern int MPI_Comm_rank(MPI_Comm, int *);
extern int MPI_Comm_dup(MPI_Comm, MPI_Comm *);
/* extern int MPI_Comm_rank(MPI_Comm, int, int, MPI_Comm*); */


/* Group functions */
extern int MPI_Group_union(MPI_Group, MPI_Group, MPI_Group *);
extern int MPI_Group_intersection(MPI_Group, MPI_Group, MPI_Group *);
extern int MPI_Group_difference(MPI_Group, MPI_Group, MPI_Group *);
extern int MPI_Group_incl(MPI_Group, int, int *, MPI_Group *);
extern int MPI_Group_excl(MPI_Group, int, int *, MPI_Group *);
extern int MPI_Group_size(MPI_Group, int *);
extern int MPI_Group_rank(MPI_Group, int *);
extern int MPI_Group_compare(MPI_Group, MPI_Group, int *);
extern int MPI_Group_free(MPI_Group *);


/* other functions */

extern int MPI_Reduce(void *, void *, int, MPI_Datatype, 
		      MPI_Op, int, MPI_Comm);
int MPI_Sizeof(MPI_Datatype);
void print_group(MPI_Group group);

#define mpi_header
#endif










